;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "annotated-completing-read" "20260817.235"
  "Ergonomic completing-read wrapper/helper."
  '((emacs "29.1"))
  :url "https://github.com/tychoish/annotated-completing-read"
  :commit "e93c3d71d59a4a01c0a28c3d4c9ecca00e5d83fc"
  :revdesc "e93c3d71d59a"
  :keywords '("convenience" "matching")
  :authors '(("sam kleinman" . "garen@tychoish.com"))
  :maintainers '(("sam kleinman" . "garen@tychoish.com")))
